{U+,R+}
program BarCode;
const
  MaxIns = 350;
  MaxCode = 512;

  Ver = 'Barcode Generator V1.7a';

  ErrA = 'Illegal ';
  ErrB = 'Missing ';

type
  charset = set of ' '..'~';
  str2  = string[2];
  str3  = string[3];
  str16 = string[16];
  InsType = record
    Str : string[8];
    Class,
    B1,
    B2 : byte
    end;

const
  Digits : charset = ['0'..'9','E','-','.'];

var
  F : text;             {I/O var}
  FN : string[14];
  S,Line : string[80];  {HP-41 instruction to compile}
  Title : string[10];
  P : integer;          {Position variable}
  Arg : byte;
  LineNo : integer;
  Errors : integer;
  Ins : array[0..MaxIns] of InsType;
  IR : InsType;
  NumIns : integer;
  Status : array[0..MaxCode] of boolean;
  Code : array[0..MaxCode] of byte;
  PC : 0..MaxCode;

  LastDigit : boolean;
  Private : byte;
  Dots : array[0..960] of char;
  DP : 0..960;
  Bytes : array[0..15] of byte;
  BP : 0..16;
  RowNo, BLN : integer;
  XSum : integer;
  ch : char;

function Nyble(N : byte) : char;
begin
  if N>9 then
    N:=N+7;
  Nyble:=chr(48+N)
  end;

function Hex(H : byte) : str2;
begin
  Hex:=Nyble(H div 16)+Nyble(H mod 16)
  end;

function FileOK(S : str3) : boolean;
begin
  if FN='' then
    FileOK:=false
  else begin
    FileOK:=true;
    if Pos(FN,'.')=0 then
      FN:=FN+'.'+S;
    assign(F,FN);
    {$I-}
    reset(F);
    {$I+}
    if IOresult<>0 then begin
      writeln('File not found');
      FileOK:=false
      end
    end
  end;

procedure ReadFile;
begin
  repeat
    NumIns:=succ(NumIns);
    if NumIns>MaxIns then begin
      writeln('Out of memory.');
      halt
      end;
    with Ins[NumIns] do begin
      read(F,Class,B1);
      if Class in [4,5,7] then
        read(F,B2);               {Classes 4,5,7 are two bytes}
      readln(F,Str)
      end
    until Ins[NumIns].Class in [255,254];
  if Ins[NumIns].Class=254 then
    NumIns:=pred(NumIns);
  writeln(NumIns,' definitions read');
  close(F);
  end;

procedure ReadIns;
begin
  NumIns:=0;
  FN:='HP41';
  if FileOK('INS') then
    ReadFile
  else
    halt;
  writeln('enter XROM files:');
  repeat
    writeln;
    write('-');
    readln(FN);
    if FileOK('INS') then
      Readfile
    until FN='';
  end;

procedure GetChar;
begin
  ch:=S[1];
  delete(S,1,1)
  end;

procedure SOB;
begin
  while S[1]=' ' do
    GetChar
  end;

function CharConv(C : char) : byte;
begin
  if C='&' then
    CharConv:=126        {Sigma}
  else if C='|' then
    CharConv:=127        {Append}
  else if C='@' then
    CharConv:=13         {Angle sign}
  else
    CharConv:=ord(C)
  end;

procedure Err(E : integer);
begin
  writeln;
  writeln(LineNo:3,'  ',Line);
  Errors:=succ(Errors);
  case E of
    1 : write(ErrA,'function');
    2 : write(ErrA,'argument(s)');
    3 : write(ErrA,'#');
    4 : write(ErrB,'''"''');
    5 : write(ErrB,''')''')
    end;
  writeln
  end;

procedure Stuff(B : byte);
begin
  Code[PC]:=B;
  PC:=succ(PC);
  end;

procedure Mark(N,S : integer);
begin
  FillChar(Dots[DP],N,255);
  DP:=DP+S+N
  end;

procedure One;
begin
  Mark(5,3);
  end;

procedure Zero;
begin
  Mark(2,3)
  end;

procedure ConvByte(B : byte);
var
  I : integer;

begin
  for I:=0 to 7 do
    if B shl I and 128=0 then     {Check MSB}
      Zero
    else
      One
  end;

procedure PrintRow;
var
  I : integer;

  procedure PR;
  begin
    write(Lst,chr(27),'L',chr(succ(DP) mod 256),chr(succ(DP) div 256));
    for I:=0 to DP do
      write(lst,Dots[I]);
    end;

begin  {procedure PrintRow}
  writeln(Lst);
  write(Lst,^O,Title,'':3,'Row ',RowNo,' (',BLN,'-',LineNo,')','':3);
  FillChar(Dots,sizeof(Dots),0);  {Clear out dots array}
  DP:=0;
  Zero;  Zero;               {Put in start bits}
  for I:=1 to BP do          {Calculate checksum}
    XSum:=XSum+Bytes[I];
  XSum:=XSum mod 255;
  if XSum=0 then
    XSum:=255;
  ConvByte(XSum);            {Put barcode in dots array}
  write(lst,Hex(XSum));      {Write a hex representation}
  for I:=1 to BP do begin
    ConvByte(Bytes[I]);
    write(Lst,':',Hex(Bytes[I]))
    end;
  writeln(lst);
  One;  Zero;                {Put in stop bits}
  PR;
  write(Lst,chr(27),'J',chr(24),^M);    {Advance paper}
  PR;
  writeln(Lst)
  end;

function GetString : str16;
begin
  GetChar;
  P:=Pos('"',S);
  if P=0 then begin
    Err(4);
    P:=1
    end;
  GetString:=copy(S,1,pred(P))
  end;

procedure PutString(S : str16);
var
  i : integer;

begin
  stuff(240+length(S));
  for I:=1 to length(S) do
    stuff(CharConv(S[I]))
  end;

function IND : boolean;
begin
  IND:=false;
  if copy(S,1,4)='IND ' then begin
    IND:=true;
    Delete(S,1,4);
    SOB
    end
  end;

procedure GetArg;
var
  IndF : boolean;

begin
  Arg:=0;
  IndF:=IND;
  GetChar;
  if ch in ['0'..'9'] then
    while ch in ['0'..'9'] do begin
      Arg:=Arg*10+ord(ch)-48;
      GetChar
      end
  else begin
    Arg:=pos(ch,'ABCDEFGHIJTZYXLMNOPQ|abcde')+101;
    if Arg=101 then
      Err(2)
    end;
  if IndF then
    Arg:=Arg+128;
  end;

procedure PutXROM(I,J : integer);
begin
  Stuff(160+((I and 63) shr 2));
  Stuff(((I and 3) shl 6)+J and 63)
  end;

procedure CkNumber;
label
  Exit;

var
  LastPC : integer;

begin
  LastPC:=PC;
  if LastDigit then
    Stuff(0);
  LastDigit:=false;
  P:=1;
  repeat
    ch:=S[P];
    P:=succ(P);
    if ch in ['0'..'9','.','E'] then
      LastDigit:=true;
    case ch of
      '0'..'9'  : Stuff(ord(ch)-32);
      '.'       : Stuff(26);
      'E'       : Stuff(27);
      '-'       : Stuff(28);
      ' '       : ;
    else
      LastDigit:=false;
      Goto Exit;
      end
    until (ch=' ') and (S[P]<>'E');
Exit:
  if not LastDigit then
    PC:=LastPC;
  end;

procedure DoSpecial;
var
  N : integer;
  Flag : boolean;

begin
  N:=0;
  Flag:=false;
  GetChar;
  SOB;
  repeat
    GetChar;
    case ch of
      ' ',')'   : begin
                    if Flag then
                      Stuff(N);
                    N:=0;
                    Flag:=false
                    end;
    '0'..'9',
    'A'..'F'  : begin
                   Flag:=true;
                   N:=N*16+ord(ch)-48;
                   if ch in ['A'..'F'] then
                     N:=N-7
                   end;
    else
                  err(3);
                  ch:=')'
                  end;
    until (ch=')') or (S='');
  if ch<>')' then
    Err(5)
  end;

procedure DoEnd;
begin
  Stuff(192);
  Stuff(0);
  Stuff(47);
  Status[PC]:=true
  end;

procedure DoLBL;
begin
  If S[1]='"' then begin
    Stuff(192);                   {Stuff a Global Chain}
    Stuff(0);
    PutString(chr(0)+GetString)
    end
  else begin
    GetArg;
    if Arg<15 then
      Stuff(succ(Arg))            {Use a short-form label}
    else begin
      Stuff(207);
      Stuff(Arg)
      end
    end
  end;

procedure DoGtoXeq;
var
  GTO : boolean;

begin
  GTO:=IR.Str='GTO';
  If S[1]='"' then begin
     Stuff(IR.B1);
     PutString(GetString)
     end
   else if IND then begin
     Stuff(174);
     GetArg;
     if GTO then
       Stuff(Arg)
     else
       Stuff(128+Arg)
     end
   else begin
     GetArg;
     if GTO and (Arg<15) then begin
       Stuff(177+Arg);
       Stuff(0)
       end
     else begin
       Stuff(IR.B2);
       Stuff(0);
       Stuff(Arg)
       end
     end
  end;

procedure DoStoRcl;
begin
  GetArg;
  if Arg<16 then
    Stuff(Arg+IR.B1)
  else begin
    Stuff(IR.B2);
    Stuff(Arg)
    end
  end;

procedure DoTable;  forward;

procedure DoXROM;
var
  T : integer;

begin
  if S[1]='"' then
    DoTable
  else begin
    GetArg;
    T:=Arg;
    if ch<>',' then
      Err(2)
    else begin
      GetChar;
      GetArg;
      PutXrom(T,Arg)
      end
    end
  end;

procedure DoTable;
var
  i : integer;
  InsStr : string[8];

begin
  P:=pos(' ',S);
  InsStr:=copy(S,1,pred(P));
  Delete(S,1,P);
  SOB;
  I:=1;
  while (I<=NumIns) and (InsStr<>Ins[I].Str) do
    I:=succ(I);
  if I>NumIns then
    Err(1)
  else begin
    IR:=Ins[I];
    case IR.Class of
      1   : Stuff(IR.B1);
      2   : begin
            Stuff(IR.B1);
            GetArg;
            Stuff(Arg)
            end;
      3   : DoLBL;
      4   : DoStoRcl;
      5   : DoGtoXeq;
      6   : DoXROM;
      7   : PutXROM(IR.B1,IR.B2);
      255 : DoEnd;
      else
        Err(99)
      end
    end
  end;

procedure PRS;
begin
  FillChar(Code,sizeof(Code),0);
  FillChar(Status,sizeof(Status),0);
  FillChar(IR,sizeof(IR),0);
  PC:=0;
  Errors:=0;
  LineNo:=0;
  XSum:=0;
  repeat
    writeln;
    write('*');
    readln(FN);
    if FN='' then halt;
    until FileOK('41');
  end;

procedure Compile;
begin
  repeat
    readln(F,Line);
    if (Line<>'') and (Line[1]<>'\') then begin
      S:=Line+' $';
      if LineNo mod 8=0 then
        write(LineNo:3,^M);
      LineNo:=succ(LineNo);
      Status[PC]:=true;      {Mark new instruction}
      CkNumber;
      if not LastDigit then       {Command not a number}
        case S[1] of
          '"' : PutString(GetString);
          '(' : DoSpecial
          else
                DoTable
                end
      end
    until IR.Class=255;
  writeln(LineNo:3);
  end;

procedure PrintBarCode;
var
  I : integer;

begin
  writeln(Lst);
  BP:=0;                {# of bytes in row}
  RowNo:=1;             {Barcode row #}
  LineNo:=0;
  For I:=0 to pred(PC) do begin
    if Status[I] then
      LineNo:=succ(LineNo);
    if BP=0 then begin            {Set up header bytes}
      Bytes[1]:=Private+ pred(RowNo) mod 16;
      BLN:=LineNo;
      while not Status[BP+I] do
        BP:=succ(BP);             {Broken bytes at beginning}
      Bytes[2]:=BP shl 4;
      BP:=2;                 {Set up byte pointer}
      end;
    BP:=succ(BP);
    Bytes[BP]:=Code[I];
    if BP=15 then begin
      BP:=0;
      while not Status[succ(I)-BP] do
        BP:=succ(BP);
      Bytes[2]:=Bytes[2]+BP;
      BP:=15;
      PrintRow;
      BP:=0;
      if RowNo mod 16=0 then
        write(Lst,^L);
      RowNo:=succ(RowNo)
      end
    end;
  if BP<>0 then
    PrintRow;
  writeln(Lst);
  writeln(Lst,'':90,PC,' bytes used (',succ(pred(PC) div 7),' registers)');
  end;

begin
  writeln(Ver);
  ReadIns;
    repeat
    PRS;
    Compile;
    writeln;
    writeln(PC,' bytes used.');
    writeln;
    if Errors=0 then begin
      write('Do you wish to print the barcode? ');
      readln(ch);
      if upcase(ch)='Y' then begin
        write('Enter the name of the program: ');
        readln(Title);
        write('Is the program private? ');
        readln(ch);
        Private:=16;
        if upcase(ch)='Y' then
          Private:=32;
        PrintBarCode
        end
      end
    else
      writeln(Errors, ' error(s) found.')
    until false;
  end.

