{$U+           control-C works}
program FSA;
const
  Null = '@';      {Lamda in Regular Expressions}
  QSize = 50;

type
  NP = ^NextPtr;
  SP = ^State;
  State = record
    Match : char;
    Next : NP
    end;
  NextPtr = record
    NewState : SP;
    Next : NP;
    end;
  Str40 = string[40];
  QPtr = 0..Qsize;

var
  Start : ^State;
  Inp : Str40;           {String to check against}
  Rule : Str40;
  Queue : array[QPtr] of ^State;
  Head, Tail : Qptr;

procedure PRS;
var
  I : QPtr;

begin
  for I:=0 to Qsize do
    Queue[I]:=nil;
  Head:=0;
  Tail:=0
  end;

function Qadd(I : integer) : QPtr;
begin
  while I<=0 do
    I:=I+Qsize;
  Qadd:=I mod Qsize
  end;

procedure CheckFull;
begin
  if Head=Tail then begin
    writeln('Queue overflow!');
    halt
    end
  end;

procedure CheckMT;
begin
  if Head=Tail then begin
    writeln('Queue underflow!');
    halt
    end
  end;

procedure AddEnd(S : SP);
begin
  Queue[Tail]:=S;
  Tail:=Qadd(succ(Tail));
  CheckFull;
  end;

procedure AddTop(S : SP);
begin
  Head:=Qadd(pred(Head));
  CheckFull;
  Queue[Head]:=S
  end;

function GetEnd : SP;
begin
  CheckMT;
  Tail:=Qadd(pred(Tail));
  GetEnd:=Queue[Tail];
  end;

function GetTop : SP;
begin
  CheckMT;
  GetTop:=Queue[Head];
  Head:=Qadd(succ(Head));
  end;

procedure CheckInp;
var
  I : integer;
  ch : char;
  StopHead : QPtr;
  S : ^State;
  N : ^NextPtr;
  NullFlag, Term : boolean;

begin
  AddEnd(Start);
  for I:=0 to length(Inp) do begin
    if I=0 then
      ch:=Null
    else
      ch:=Inp[I];
    writeln('Pos ',I,' - ',ch);
    StopHead:=Head;
    Term:=false;
    while Tail<>StopHead do begin
      S:=GetEnd;
      writeln('':2,'>',S^.Match,' ',ord(S));
      if (S^.Match=Null) or (S^.Match=ch) then begin
        N:=S^.Next;
        if N=nil then
          Term:=true;
        while N<>nil do begin
          S:=N^.NewState;
          write('':4);
          if S^.Match=null then begin
            write('|');
            AddEnd(S)          {Do null states right away}
            end
          else begin
            write('<');
            AddTop(S);         {Defer real characters}
            end;
          writeln(S^.Match,' ',ord(S));
          N:=N^.Next
          end
        end
      end;
    if (Head=StopHead) and (not Term) then
      if I=0 then
        AddEnd(Start)            {Restart at end}
      else
        writeln('String mismatch at position ',I);
    end;
  if Term=true then
    writeln('String matches.')
  else
    writeln('Unexpected end of string.');
  end;

{-------------Here begineth the parser--------}

function NewNode : SP;
var
  S : SP;

begin
  New(S);
  S^.Match:=null;
  S^.Next:=nil;
  NewNode:=S;
  end;

procedure StuffBranch(S,NS : SP);
var
  N : NP;

begin
  new(N);
  N^.Next:=S^.Next;
  S^.Next:=N;
  N^.NewState:=NS
  end;


{$A-          recursive code begins here      }

function Compile(Rule : Str40) : SP;      {Returns pointer to evaluated
                                          expression}

const
  EOL = ^M;
  Union = '+';    {Sometimes '+'}
  Kleene = '*';
  Letters : set of char = ['A'..'Z','0'..'9'];

type
  Result = record
    B,E : SP;
    end;

var
  Pos : integer;       {Current position in the Rule}
  ErrPos : integer;
  ch : char;           {Current charater}
  R : result;

procedure NextCh;
begin
  Pos:=Pos+1;
  if Pos<=length(Rule) then
    ch:=Rule[Pos]
  else
    ch:=EOL;
  end;

procedure Expression(var R : result); forward;

procedure Factor(var R : result);

procedure GetLetter;
var
  S : sp;

begin
  write('(GL)');
  S:=NewNode;
  S^.Match:=ch;
  R.B:=S;
  R.E:=S;
  NextCh;
  end;

procedure GetParen;        {Evaluate a parenthesis group}
begin
  write('(GP)');
  NextCh;
  Expression(R);
  if Ch=')' then
    NextCh
  else
    ErrPos:=Pos;
  end;

begin {procedure Factor}
  write('(F)');
  if Ch in Letters then
    GetLetter
  else if ch='(' then
    GetParen
  else ErrPos:=Pos
  end;

procedure KleenedFactor(var R : result);
var
  S,T : SP;

begin
  write('(KF)');
  Factor(R);
  if ch=Kleene then begin
    NextCh;
    S:=NewNode;           {null nodes for Kleene star processing}
    T:=NewNode;
    StuffBranch(R.E,R.B);    {loop!}
    StuffBranch(S,R.B);
    StuffBranch(S,T);        {if kleene is not executed at all}
    StuffBranch(R.E,T);      {provide new end node}
    R.B:=S;
    R.E:=T
    end
  end;

procedure SimpleExpression(var R : result);
var
  Q : result;

begin
  write('(SE)');
  KleenedFactor(R);
  while (ch in Letters) or (ch='(') do begin
    KleenedFactor(Q);
    StuffBranch(R.E,Q.B);     {Concatenate two factors}
    R.E:=Q.E;
    end
  end;

procedure Expression;
var
  Q : result;
  S,T : SP;

begin
  write('(E)');
  SimpleExpression(R);
  while ch=Union do begin
    NextCh;
    SimpleExpression(Q);
    S:=NewNode;            {set up null nodes for union}
    T:=NewNode;
    StuffBranch(S,R.B);
    StuffBranch(S,Q.B);
    StuffBranch(R.E,T);
    StuffBranch(Q.E,T);
    R.B:=S;
    R.E:=T
    end
  end;

begin {procedure Compile}
  ErrPos:=0;
  Pos:=0;
  NextCh;
  Expression(R);
  if ch<>EOL then begin
    writeln;
    repeat
      write(ch);
      nextch;
      until ch=EOL;
    writeln(' <---- Extra ignored');
    end;
  if ErrPos<>0 then begin
    writeln('Error in rule compile at position ',ErrPos);
    halt
    end;
  Compile:=R.B;
  end;

{$A+}

procedure Debug;
var
  Add : integer;
  S : SP;
  N : NP;

begin
  writeln('Start = ',ord(Start));
  writeln;
  repeat
    write('Address: ');
    readln(Add);
    S:=ptr(Add);
    writeln('Match - ',S^.Match);
    writeln('Links:');
    N:=S^.Next;
    while N<>nil do begin
      writeln('':4,ord(N^.NewState));
      N:=N^.Next
      end;
    writeln
    until Add=0
  end;

begin
  write('Enter Rule: ');
  readln(Rule);
  Start:=compile(Rule);
  writeln;
  repeat
    write('Enter string: ');
    readln(Inp);
    PRS;
    CheckInp;
    until Inp='';
  end.
